"""Utils file."""

from click.testing import CliRunner
from teledetection.sdk.logger import get_logger_for


log = get_logger_for(__name__)


def run_cli_cmd(command, args=None):
    """Test a CLI command."""
    log.info("Testing %s", command)
    runner = CliRunner()
    result = runner.invoke(command, args)
    log.info("Result is %s", result)
    assert result.exit_code == 0


def should_fail(func, args, exception_cls):
    """Helper with intended failing tests."""
    try:
        if isinstance(args, list):
            func(*args)
        else:
            func(**args)
        raise AssertionError("This test should fail!")
    except exception_cls:
        print("Failing as expected :)")



"""Push test module."""

import os
import time
import datetime
import requests
import pystac_client

from utils import run_cli_cmd, should_fail

import teledetection
from teledetection.cli import (
    _create_new_key,
    _get_all_keys,
    apikey,
)
from teledetection.sdk.logger import get_logger_for


STAC_ENDPOINT = "https://api.stac.teledetection.fr"

log = get_logger_for(__name__)


def _search(collection, bbox, asset_key):
    """STAC search."""
    start = time.time()
    api = pystac_client.Client.open(
        STAC_ENDPOINT,
        modifier=teledetection.sign_inplace,
    )
    res = api.search(
        bbox=bbox,
        datetime=["2017-01-01", "2022-12-31"],
        collections=[collection],
    )
    nb = 100
    for i, item in enumerate(res.items()):
        log.info(i)
        if i == nb:
            break
        url = item.assets[asset_key].href
        assert "Amz" in url

    elapsed = time.time() - start
    log.info(f"Took {round(elapsed, 2)} s")

    response = requests.get(url, timeout=5, stream=True)
    response.raise_for_status()


def _spot67():
    """Test spot-6/7 collection."""
    _search(bbox=[-3.75, 30, 10, 60], collection="spot-6-7-drs", asset_key="src_xs")


def _super_s2():
    """Test Super-S2 collection."""
    _search(
        bbox=[3.75, 43.58, 3.95, 43.67],
        collection="super-sentinel-2-l2a",
        asset_key="img",
    )


def _misc():
    """Various tests."""

    def test_userinfo():
        """Test userinfo method."""
        log.info("User infos: %s", teledetection.get_userinfo())

    def test_username():
        """Test userinfo method."""
        log.info("Username: %s", teledetection.get_username())

    test_userinfo()
    test_username()


def _headers(key: str):
    """Get headers test."""
    headers = teledetection.get_headers()
    log.info("Got headers")
    assert headers
    assert headers.get(key)


def _perform(key: str):
    """Perform tests."""
    _headers(key)
    _misc()
    _super_s2()
    _spot67()


def _oauth2_needing_reauth():
    """Some tests that need OAuth2 reauth."""
    # Save a JTW
    oauth2 = teledetection.sdk.oauth2
    good_jwt = teledetection.sdk.model.JWT.from_config_dir()

    # Screw saved JTW (expiry dates)
    jwt = good_jwt.model_copy()
    jwt.expires_in = 0
    jwt.to_config_dir()
    time.sleep(1)
    assert oauth2.OAuth2Session().get_access_token()

    # Screw saved JTW (tokens)
    jwt = good_jwt.model_copy()
    jwt.access_token += "corrupted"
    jwt.refresh_token += "corrupted"
    jwt.to_config_dir()
    assert oauth2.OAuth2Session().get_access_token()

    log.warning(
        """
        ****************************************************************************

        Simulate an expired authentication !
        You developer will have to wait! sorry...

        DO NOT LOGIN PLEASE, THIS TEST IS PURPOSELY INTENDED TO FAIL

        ****************************************************************************
        """
    )
    right_cfg_dir = teledetection.sdk.settings.get_config_path()
    cfg_dir_cpy = f"{right_cfg_dir}_old"
    os.rename(right_cfg_dir, cfg_dir_cpy)
    oauth2_sess = oauth2.OAuth2Session()
    oauth2_sess.jwt = None
    should_fail(oauth2_sess.get_access_token, [], oauth2.ExpiredAuthLinkError)
    os.rename(cfg_dir_cpy, right_cfg_dir)


def test_sdk():
    """Test SDK auth with various authentication methods."""
    # QR-code will be shown there. An OAuth2 token will be generated.
    log.info("Testing key management")

    # Revoke all keys from CLI
    run_cli_cmd(apikey, ["revoke-all"])
    assert len(_get_all_keys()) == 0

    # Test CLI's list with 0 key
    run_cli_cmd(apikey, ["list"])

    # Create new key from CLI
    run_cli_cmd(apikey, ["create"])
    keys = _get_all_keys()
    assert isinstance(keys, list)
    assert len(keys) == 1
    run_cli_cmd(apikey, ["create", "my super key123"])
    keys = _get_all_keys()
    assert isinstance(keys, list)
    assert len(keys) == 2

    # Revoke a key from CLI
    key = _create_new_key(description="key to be deleted")
    access_key = key["access-key"]
    run_cli_cmd(apikey, ["revoke", access_key])
    assert access_key not in _get_all_keys()

    # Register key with CLI
    run_cli_cmd(apikey, ["register"])

    # Remove key with CLI
    run_cli_cmd(apikey, ["remove"])

    # List keys with CLI
    run_cli_cmd(apikey, ["list"])

    log.info("Testing SDK with OAuth2")

    # Apply configuration (none here, because we have removed all the
    # API keys, so now we only rely on the OAuth2 JWT token)
    teledetection.sdk.http.session.prepare_connection_method()

    # Tests and expect authorization bearer in headers
    _perform("authorization")

    log.info("Test SDK with API Key from env. var.")

    # Create a key, get the key as dict
    key = _create_new_key(description="")
    assert key

    # Put the key in env. var.
    os.environ["TLD_ACCESS_KEY"] = key["access-key"]
    os.environ["TLD_SECRET_KEY"] = key["secret-key"]

    # Ensure environment variables are used
    teledetection.sdk.http.session.prepare_connection_method()

    # Tests and expect access key in headers
    _perform("access-key")

    # Remove key
    run_cli_cmd(apikey, ["revoke", key["access-key"]])
    os.environ.pop("TLD_ACCESS_KEY")
    os.environ.pop("TLD_SECRET_KEY")

    log.info("Testing SDK with API key")

    # Create a new API key stored in the .config folder
    run_cli_cmd(apikey, ["register"])

    # Ensure that the API key is used, even when the .jwt
    # is here after the OAuth2 connection.
    teledetection.sdk.http.session.prepare_connection_method()

    # Tests and expect access key in headers
    _perform("access-key")

    # Delete locally stored key
    run_cli_cmd(apikey, ["remove"])

    # Switch back to OAuth2
    teledetection.sdk.http.session.prepare_connection_method()

    if os.environ.get("SKIP_AUTH_LINK") not in ("on", "true", "ON", "TRUE", "1"):
        _oauth2_needing_reauth()


def test_cfg_dir():
    """Test config dir."""
    assert os.path.exists(teledetection.sdk.settings.get_config_path())


def test_http():
    """Test http module."""
    http = teledetection.sdk.http
    assert isinstance(http.BareConnectionMethod().get_headers(), dict)
    assert not http.BareConnectionMethod().get_headers()

    assert teledetection.sdk.http.get_headers()
    assert teledetection.sdk.http.get_userinfo()
    assert teledetection.sdk.http.get_username()

    # Revoke all API keys
    run_cli_cmd(apikey, ["revoke-all"])

    http_sess = http.HTTPSession()
    http.ENV.tld_disable_auth = True
    assert isinstance(http_sess.get_method(), http.BareConnectionMethod)

    http_sess = http.HTTPSession()
    http.ENV.tld_disable_auth = False
    assert isinstance(http_sess.get_method(), http.OAuth2ConnectionMethod)

    should_fail(
        http_sess.post,
        {"route": "nonexistent-route", "params": {}},
        requests.exceptions.HTTPError,
    )


def test_should_fail():
    """Test should_fail function."""

    def no_error():
        print("yo")

    try:
        should_fail(no_error, [], ValueError)
    except AssertionError:
        # Message "this test should fail"
        pass


def test_model():
    """Model tests."""
    model = teledetection.sdk.model
    file = "/non-existent-dir/file.json"
    model.Serializable().to_file(file)


def test_oauth2():
    """OAuth2 tests."""
    oauth2 = teledetection.sdk.oauth2

    # Not implemented error
    gmb = oauth2.GrantMethodBase()
    gmb.client_id = "client"
    should_fail(gmb.get_first_token, [], NotImplementedError)

    # Simulate a failure in refresh token
    teledetection.sdk.utils.ENV.tld_retry_total = 2
    jwt = teledetection.sdk.model.JWT(
        access_token="",
        expires_in=0,
        refresh_token="",
        refresh_expires_in=0,
        token_type="",
    )
    should_fail(gmb.refresh_token, [jwt], oauth2.RefreshTokenError)

    # Test working refresh token
    oauth2_sess = oauth2.OAuth2Session()
    oauth2_sess._refresh_if_needed()  # pylint: disable=protected-access

    # Simulate a failure in get first token
    oauth2_sess.grant.client_id = "not-existing-client-id"
    oauth2_sess.grant._token_endpoint = None  # pylint: disable=protected-access
    should_fail(oauth2_sess.grant.get_first_token, [], ConnectionError)

    # Save no jwt
    oauth2.OAuth2Session().save_token(datetime.datetime.now())


def _check_signed(s: str):
    """Check that s contains signature."""
    assert all(
        word in s
        for word in (
            "X-Amz-Algorithm",
            "X-Amz-Credential",
            "X-Amz-Date",
            "X-Amz-Expires",
            "X-Amz-SignedHeaders",
            "X-Amz-Signature",
        )
    )


def test_sign():
    """Test signing module."""
    should_fail(teledetection.sign, [1234], TypeError)
    assert teledetection.sdk.signing.is_vrt_string("<VRTDataset...</VRTDataset>")

    # Get an asset href
    api = pystac_client.Client.open(
        STAC_ENDPOINT,
        modifier=teledetection.sign_inplace,
    )
    res = api.search()
    item = next(res.items())
    href = item.assets[list(item.assets.keys())[0]].href

    # Asset signing (copy)
    for asset in item.assets.values():
        asset_clone = teledetection.sign(asset)
        _check_signed(asset_clone.href)

    # Item signing (copy)
    item_clone = teledetection.sign(item)
    for asset in item_clone.assets.values():
        _check_signed(asset.href)

    # Item signing (in place)
    teledetection.sign_inplace(item)
    for asset in item.assets.values():
        _check_signed(asset.href)

    # Asset signing (in place)
    item = next(res.items())
    for asset in item.assets.values():
        teledetection.sign_inplace(asset)
        _check_signed(asset.href)

    # URL signing
    _check_signed(teledetection.sign(href))

    # VRT string signing
    vrt_like = f"<VRTDataset>...dummy VRT here...{href}</VRTDataset>"
    _check_signed(teledetection.sign(vrt_like))

    # URL outside domain signing
    url_out = "https://something-else-outside.org/file.tif"
    assert teledetection.sign(url_out) == url_out



"""Test file."""

import os
import glob
import shutil
import tempfile
import urllib.request
from datetime import datetime

import pystac
import pystac_client
import pytest
import requests

from utils import run_cli_cmd


from teledetection import sign, sign_inplace
from teledetection.upload import raster, stac, diff, transfer
from teledetection.cli import (
    collection_diff,
    grab,
    list_col_items,
    list_cols,
    publish,
    DEFAULT_S3_EP,
)
from teledetection.sdk.logger import get_logger_for


log = get_logger_for(__name__)

STAC_ENDPOINT = "https://api-dev.stac.teledetection.fr"
DEFAULT_COL_HREF = "http://hello.fr/collections/collection-for-tests"
IMAGE_HREF = (
    "https://gitlab.orfeo-toolbox.org/orfeotoolbox/"
    "otb/-/raw/develop/Data/Input/Capitole_Rasterization.tif"
)
COL_ID = "test-collection-for-upload"
items_ids = ["item_1", "item_2"]
RASTER_FILE1 = "/tmp/raster1.tif"
RASTER_FILE2 = "/tmp/folder1/raster2.tif"
RASTER_FILE3 = "/tmp/folder/raster3.tif"
STORAGE_BUCKET = "sm1-gdc-tests"

handler = stac.StacUploadTransactionsHandler(
    stac_endpoint=STAC_ENDPOINT,
    storage_endpoint=DEFAULT_S3_EP,
    storage_bucket=STORAGE_BUCKET,
    assets_overwrite=True,
    sign=False,
)

with open(RASTER_FILE1, "wb") as f:
    r = requests.get(IMAGE_HREF, timeout=5)
    f.write(r.content)
os.makedirs(os.path.dirname(RASTER_FILE2), exist_ok=True)
shutil.copyfile(RASTER_FILE1, RASTER_FILE2)
os.makedirs(os.path.dirname(RASTER_FILE3), exist_ok=True)
shutil.copyfile(RASTER_FILE1, RASTER_FILE3)

COL_BBOX = [0.0, 0.0, 0.0, 0.0]
BBOX_ALL = [
    3.6962018175925073,
    43.547450099338604,
    9.036414917971516,
    48.75431706444037,
]
COORDS1 = [
    [4.032730583418401, 43.547450099338604],
    [4.036414917971517, 43.75162726634343],
    [3.698685718905037, 43.75431706444037],
    [3.6962018175925073, 43.55012996681564],
    [4.032730583418401, 43.547450099338604],
]
COORDS2 = [[coord + 5 for coord in coords] for coords in COORDS1]


def clear():
    """Clear all test items and collection."""
    for item_id in items_ids:
        handler.delete_item_or_col(col_id=COL_ID, item_id=item_id)
    handler.delete_item_or_col(col_id=COL_ID)


def remote_col_test(expected_bbox):
    """Run tests on a remote collection."""
    api = pystac_client.Client.open(STAC_ENDPOINT)
    col = api.get_collection(COL_ID)
    extent = col.extent.spatial.bboxes
    assert len(extent) == 1
    assert tuple(extent[0]) == tuple(expected_bbox), (
        f"expected BBOX: {expected_bbox}, got {extent[0]}"
    )

    # Check that assets are accessible once signed
    for i in col.get_items():
        assets = i.get_assets().values()
        for asset in assets:
            assert stac.asset_exists(asset.href)
            assert "?" not in asset.href, f"The asset URL looks signed: {asset.href}"
            sign_inplace(asset)
            urllib.request.urlretrieve(asset.href, "/tmp/a.tif")
            assert raster.is_cog("/tmp/a.tif")


def create_item(item_id: str):
    """Create a STAC item."""
    coordinates = COORDS1 if item_id == "item_1" else COORDS2

    # Check is COG
    assert not raster.is_cog(RASTER_FILE1)
    assert not raster.is_cog(RASTER_FILE2)
    assert not raster.is_cog(RASTER_FILE3)

    item = pystac.Item(
        id=item_id,
        geometry={
            "type": "Polygon",
            "coordinates": [coordinates],
        },
        bbox=[
            min(coords[0] for coords in coordinates),
            min(coords[1] for coords in coordinates),
            max(coords[0] for coords in coordinates),
            max(coords[1] for coords in coordinates),
        ],
        datetime=datetime.now().replace(year=1999),
        properties={},
        assets={
            "ndvi": pystac.Asset(href=RASTER_FILE1),
            "crswir": pystac.Asset(href=RASTER_FILE2),
            "ndwi": pystac.Asset(href=RASTER_FILE3),
        },
    )

    return item


def create_collection(col_href: str):
    """Create an empty STAC collection."""
    spat_extent = pystac.SpatialExtent([COL_BBOX])
    temp_extent = pystac.TemporalExtent(intervals=[[None, None]])  # type: ignore
    col = pystac.Collection(
        id=COL_ID,
        extent=pystac.Extent(spat_extent, temp_extent),
        description="Some description",
        title="Dummy collection for teledetection-upload tests",
        href=col_href,
        providers=[
            pystac.Provider("INRAE"),
        ],
    )
    return col


def create_items_and_collection(relative, items=None, col_href=DEFAULT_COL_HREF):
    """Create two STAC items attached to one collection."""
    # Create items
    items = items or [create_item(item_id=item_id) for item_id in items_ids]

    # Attach items to collection
    col = create_collection(col_href)
    for item in items:
        col.add_item(item)
    if relative:
        col.make_all_asset_hrefs_relative()
    else:
        col.make_all_asset_hrefs_absolute()

    return col, items


def generate_collection(root_dir, relative=True, items=None):
    """Generate and save a STAC collection in {root_dir}/collection.json."""
    col, _ = create_items_and_collection(relative, items=items)
    col.normalize_hrefs(root_dir)
    col.save(
        catalog_type=pystac.CatalogType.RELATIVE_PUBLISHED
        if relative
        else pystac.CatalogType.ABSOLUTE_PUBLISHED
    )


def generate_item_collection(file_pth, relative=True):
    """Generate and save a STAC item_collection in {file_pth}."""
    _, items = create_items_and_collection(relative)
    icol = pystac.item_collection.ItemCollection(items=items)
    icol.save_object(file_pth)


def test_push():
    """Push a file remotely."""
    local_file = "/tmp/toto.txt"

    with open(local_file, "w", encoding="utf-8") as file_handle:
        file_handle.write("hello world")

    target_url = "https://s3-data.meso.umontpellier.fr/sm1-gdc-tests/titi.txt"

    transfer.push(local_filename=local_file, target_url=target_url)
    log.info("push OK")

    signed_url = sign(target_url)
    log.info("sign OK")

    res = requests.get(signed_url, stream=True, timeout=10)
    assert res.status_code == 200, "Get NOK"
    log.info("get OK")

    log.info("Done")


@pytest.mark.parametrize("overwrite", [["-o"], []])
@pytest.mark.parametrize("keep_cog_dir", [True, False])
def test_publish_w_cli(overwrite: list, keep_cog_dir: bool):
    """Handle the STAC object with CLI."""
    log.info("Testing CLI with ovr=%s and keep_cog_dir=%s", overwrite, keep_cog_dir)
    cog_dir = "/tmp/cog"
    with tempfile.TemporaryDirectory() as tmpdir:
        generate_collection(tmpdir)
        run_cli_cmd(
            publish,
            [
                os.path.join(tmpdir, "collection.json"),
                "--stac_endpoint",
                STAC_ENDPOINT,
                "--storage_endpoint",
                DEFAULT_S3_EP,
                "-b",
                STORAGE_BUCKET,
            ]
            + overwrite
            + (["--keep_cog_dir", cog_dir] if keep_cog_dir else []),
        )
        if keep_cog_dir and overwrite:
            files = glob.glob(f"{cog_dir}/*.tif")
            assert files
            shutil.rmtree(cog_dir)
        remote_col_test(BBOX_ALL)
        clear()


@pytest.mark.parametrize("assets_overwrite", [False, True])
@pytest.mark.parametrize("relative", [False, True])
def test_item_collection(assets_overwrite, relative):
    """Test item collection."""
    handler.assets_overwrite = assets_overwrite
    # we need to create an empty collection before
    col = create_collection(DEFAULT_COL_HREF)
    handler.publish_collection(col)

    with tempfile.NamedTemporaryFile() as tmp:
        generate_item_collection(tmp.name, relative=relative)
        handler.load_and_publish(tmp.name)
        remote_col_test(BBOX_ALL)
        clear()


@pytest.mark.parametrize("assets_overwrite", [False, True])
@pytest.mark.parametrize("relative", [False, True])
def test_collection(assets_overwrite, relative):
    """Test collection."""
    handler.assets_overwrite = assets_overwrite
    with tempfile.TemporaryDirectory() as tmpdir:
        generate_collection(tmpdir, relative=relative)
        col_pth = os.path.join(tmpdir, "collection.json")
        handler.load_and_publish(col_pth)
        remote_col_test(BBOX_ALL)
        clear()


@pytest.mark.parametrize("assets_overwrite", [False, True])
@pytest.mark.parametrize("relative", [False, True])
def test_collection_multipart(assets_overwrite, relative):
    """Test collection."""
    log.info(f"\nRelative: {relative}")
    handler.assets_overwrite = assets_overwrite
    for item_id in items_ids:
        with tempfile.TemporaryDirectory() as tmpdir:
            generate_collection(tmpdir, relative=relative, items=[create_item(item_id)])
            col_pth = os.path.join(tmpdir, "collection.json")
            handler.load_and_publish(col_pth)
    remote_col_test(BBOX_ALL)
    clear()


def test_get():
    """Test get collections, items, etc."""

    for col in handler.client.get_collections():
        col_from_handler = handler.client.get_collection(col.id)
        assert col_from_handler
        assert isinstance(col_from_handler, pystac.Collection)
        items = handler.get_items(col_id=col.id, max_items=1)  # test get_items()
        assert isinstance(items, list)
        for item in items:
            item_from_handler = handler.get_item(
                col_id=col.id, item_id=item.id
            )  # test get_item()
            assert item_from_handler
            assert isinstance(item_from_handler, pystac.Item)


def test_diff():
    """Test diff."""

    col1, items = create_items_and_collection(
        relative=True, col_href="/tmp/collection.json"
    )
    col2 = col1.full_copy()

    item = items[0].full_copy()
    item.id += "_test"
    col2.add_item(item, item.id)

    item = items[0].full_copy()
    item.id += "_test_other"
    col1.add_item(item, item.id)

    diff.generate_items_diff(col1, col2)
    diff.collections_defs_are_different(col1, col2)

    col1_filepath = "/tmp/col1.json"
    col1.set_self_href(col1_filepath)
    col1.save(catalog_type=pystac.CatalogType.RELATIVE_PUBLISHED)

    diff.compare_local_and_upstream(
        stac.StacTransactionsHandler(stac.DEFAULT_STAC_EP, sign=False),
        col1_filepath,
        "costarica-sentinel-2-l3-seasonal-spectral-indices-M",
    )


def test_cli_list_cols():
    """Test list cols."""
    run_cli_cmd(list_cols)
    run_cli_cmd(list_col_items, ["--col_id", "spot-6-7-drs"])


def test_cli_grab():
    """Test CLI grab."""
    run_cli_cmd(grab, ["--col_id", "spot-6-7-drs", "--out_json", "/tmp/col.json"])
    run_cli_cmd(
        grab, ["--col_id", "spot-6-7-drs", "--out_json", "/tmp/col.json", "--pretty"]
    )
    run_cli_cmd(
        grab,
        [
            "--col_id",
            "spot-6-7-drs",
            "--item_id",
            "SPOT7_MS_201805051026429_SPOT7_P_201805051026429_1",
            "--out_json",
            "/tmp/item.json",
        ],
    )
    run_cli_cmd(
        grab,
        [
            "--col_id",
            "spot-6-7-drs",
            "--item_id",
            "SPOT7_MS_201805051026429_SPOT7_P_201805051026429_1",
            "--out_json",
            "/tmp/item.json",
            "--pretty",
        ],
    )


def test_cli_diff():
    """Test diff."""
    run_cli_cmd(
        collection_diff, ["--remote_id", "spot-6-7-drs", "--col_path", "/tmp/col.json"]
    )



"""Settings from environment variables."""

import os
from pydantic_settings import BaseSettings
from pydantic.types import NonNegativeInt, PositiveInt, PositiveFloat
from pydantic import field_validator
import appdirs  # type: ignore
from .logger import get_logger_for

log = get_logger_for(__name__)

# Constants
APP_NAME = "teledetection"
MAX_URLS = 64
S3_STORAGE_DOMAIN = "meso.umontpellier.fr"
DEFAULT_SIGNING_ENDPOINT = "https://signing.stac.teledetection.fr"


class Settings(BaseSettings):
    """Environment variables."""

    tld_ttl_margin: NonNegativeInt = 1800
    tld_url_duration: NonNegativeInt = 0
    tld_config_dir: str = ""
    tld_access_key: str = ""
    tld_secret_key: str = ""
    tld_retry_total: PositiveInt = 10
    tld_retry_backoff_factor: PositiveFloat = 0.8
    tld_disable_auth: bool = False
    tld_signing_endpoint: str = DEFAULT_SIGNING_ENDPOINT

    @field_validator("tld_signing_endpoint", mode="after")
    @classmethod
    def val_endpoint_after(cls, val):
        """Post initialization."""
        if not val.lower().startswith(("http://", "https://")):
            raise ValueError(f"{val} must start with http[s]://")
        if not val.endswith("/"):
            val += "/"
        return val


ENV = Settings()


def get_config_path() -> str | None:
    """Get path to config directory (usually in ~/.config/)."""
    log.debug("Get config path")
    cfg_path = ENV.tld_config_dir or appdirs.user_config_dir(appname=APP_NAME)
    if not os.path.exists(cfg_path):
        try:
            os.makedirs(cfg_path)
            log.debug("Config dir created in %s", cfg_path)
        except PermissionError:
            log.warning("Unable to use config dir %s", cfg_path)
            cfg_path = None
    else:
        log.debug("Using existing config dir %s", cfg_path)
    return cfg_path



"""STAC diff tool."""

from typing import Tuple, List, cast
from pystac import Collection, Item
from rich import print
from teledetection.sdk.logger import get_logger_for

from . import stac

logger = get_logger_for(__name__)

UNIQUE_SEP = "___"


def collections_defs_are_different(col1: Collection, col2: Collection) -> bool:
    """Compute the diff between 2 STAC collections."""

    def fields_are_different(col1: Collection, col2: Collection, field_name: str):
        recursive_fields = field_name.split(".")

        f1 = col1
        f2 = col2
        for f in recursive_fields:
            f1 = getattr(f1, f)
            f2 = getattr(f2, f)

        if f1 != f2:
            logger.info(f"{field_name} is different: '{f1}' != '{f2}'")
            return True
        return False

    fields = [
        "extent.spatial.bboxes",
        "extent.temporal.intervals",
        "description",
        "id",
        "keywords",
        "license",
        "strategy",
        "providers",
        "title",
    ]
    return any(fields_are_different(col1, col2, field) for field in fields)


def generate_items_diff(
    col1: Collection, col2: Collection
) -> Tuple[List[Item], List[Item]]:
    """Compute the diff between 2 STAC collections.

    Returns:
    - list of items only in collection 1
    - list of items only in collection 2
    """

    def item_get_unique(i: Item) -> str:
        return i.id + UNIQUE_SEP + str(i.datetime.isoformat() if i.datetime else "")

    col1_ids = [item_get_unique(i) for i in col1.get_items()]
    col2_ids = [item_get_unique(i) for i in col2.get_items()]

    only_in_1 = set(col1_ids) - set(col2_ids)
    only_in_2 = set(col2_ids) - set(col1_ids)

    def unique_retrieve_info(unique: str, col: Collection) -> Item:
        id = unique.split(UNIQUE_SEP)[0]
        item = col.get_item(id)
        if not item:
            raise Exception(f"Item {id} not found")
        return item

    list_only_in_1 = [unique_retrieve_info(unique, col1) for unique in only_in_1]
    list_only_in_2 = [unique_retrieve_info(unique, col2) for unique in only_in_2]

    return list_only_in_1, list_only_in_2


def compare_local_and_upstream(
    handler: stac.StacTransactionsHandler,
    local_col_path: str,
    remote_col_id: str = "",
):
    """Compare a local and a remote collection.

    Args:
        handler (stac.StacTransactionHandler): object to handle the connection
        local_col_path (str): path to local collection path
        remote_col_id (str, optional): Remote collection identifier.
            If unset, will take the same id as the local collection
    """
    col_local = cast(Collection, stac.load_stac_obj(obj_pth=local_col_path))
    col_remote = handler.client.get_collection(remote_col_id or col_local.id)

    only_local, only_remote = generate_items_diff(col_local, col_remote)

    collections_defs_are_different(col_local, col_remote)

    print(f"Only local ({len(only_local)}):")
    print(only_local[:20])

    print(f"Only remote ({len(only_remote)}):")
    print(only_remote[:20])



"""Raster analysis and parsing module."""

import os
from datetime import datetime
from rasterio import warp, features, open as ropen, crs, errors  # type:ignore
from rio_stac.stac import bbox_to_geom, get_projection_info  # type:ignore
import numpy
import rasterio
from rio_cogeo import cog_translate, cog_validate
from pystac.asset import Asset
from pystac.item import Item
from pystac.extensions.projection import AssetProjectionExtension
from pystac.extensions.raster import RasterBand, RasterExtension, Statistics
from pystac.extensions.timestamps import ItemTimestampsExtension
from pystac.errors import ExtensionNotImplemented

from teledetection.sdk.logger import get_logger_for

logger = get_logger_for(__name__)

EPSG_4326 = crs.CRS.from_epsg(4326)  # pylint: disable=c-extension-no-member


class NoSpatialLayerException(Exception):
    """No spatial layer exception."""


class ScalesInputFormatException(Exception):
    """No spatial layer exception."""


class OffsetsInputFormatException(Exception):
    """No spatial layer exception."""


class Info:
    """Grabs raster information."""

    def __init__(self, raster_file):
        """Init Info class.

        Args:
            raster_file: str

        Returns:
            Info class

        """
        self.raster_file = raster_file
        with ropen(self.raster_file) as src:
            bbox = src.bounds
            geom = bbox_to_geom(bbox)
            # Reproject the geometry to "epsg:4326"
            geom = warp.transform_geom(src.crs, EPSG_4326, geom)
            self.bbox = features.bounds(geom)
            self.geom = bbox_to_geom(self.bbox)
            self.meta = src.meta
            self.gsd = src.res[0]
            self.proj_ext_info = get_projection_info(src)
            self.nodata = src.nodata
            self.area_or_point = src.tags().get("AREA_OR_POINT", "").lower()
            self.bands = src.indexes
        self.stats = None

    def band_info(self, band: int):
        """Get band info.

        Args:
            band: band index

        Returns:
            band metadata and band statistics

        """
        if band <= 0:
            raise ValueError('The "band" parameter value starts at 1')

        with ropen(self.raster_file) as src_dst:
            md = {
                "data_type": src_dst.dtypes[band - 1],
                "scale": src_dst.scales[band - 1],
                "offset": src_dst.offsets[band - 1],
            }
            if self.area_or_point:
                md["sampling"] = self.area_or_point

            # If the Nodata is not set we don't forward it.
            if src_dst.nodata is not None:
                if numpy.isnan(src_dst.nodata):
                    md["nodata"] = "nan"
                elif numpy.isposinf(src_dst.nodata):
                    md["nodata"] = "inf"
                elif numpy.isneginf(src_dst.nodata):
                    md["nodata"] = "-inf"
                else:
                    md["nodata"] = src_dst.nodata

            if src_dst.units[band - 1] is not None:
                md["unit"] = src_dst.units[band - 1]

            stats = {}
            try:
                if not self.stats:
                    self.stats = src_dst.stats(approx=True)
                statistics = self.stats[band - 1]
                stats.update(
                    {
                        "mean": statistics.mean,
                        "minimum": statistics.min,
                        "maximum": statistics.max,
                        "stddev": statistics.std,
                    }
                )
            except rasterio.errors.StatisticsError as e:
                logger.warning("Unable to compute relevant statistics: %s", e)

            return md, stats


def raster2cog(
    src_raster: str,
    dst_raster: str,
):
    """Convert a raster to Cloud Optimized Geotiff.

    Args:
        src_raster: source raster file path
        dst_raster: destination raster file path
    """
    profile = {
        "driver": "GTiff",
        "interleave": "pixel",
        "tiled": True,
        "blockxsize": 512,
        "blockysize": 512,
        "compress": "DEFLATE",
        "BIGTIFF": "IF_SAFER",
    }

    config = {
        "GDAL_NUM_THREADS": os.environ.get("GDAL_NUM_THREADS") or "ALL_CPUS",
        "GDAL_TIFF_INTERNAL_MASK": True,
        "GDAL_TIFF_OVR_BLOCKSIZE": "512",
    }

    # Change output COG filename extension if not .tif
    if not dst_raster.lower().endswith(".tif"):
        pre, _ = os.path.splitext(dst_raster)
        dst_raster = f"{pre}.tif"

    cog_translate(
        source=src_raster,
        dst_path=dst_raster,
        dst_kwargs=profile,
        config=config,
        web_optimized=False,
        progress_out=None,
        in_memory=False,
        allow_intermediate_compression=True,
    )


def is_cog(src_raster: str) -> bool:
    """Check if the raster is in COG format.

    Args:
        src_raster: source raster file path

    Returns:
        True if the raster is a COG, else False
    """
    is_valid, _, _ = cog_validate(src_raster, quiet=True, strict=True)
    return is_valid


def is_raster(src_filepath: str) -> bool:
    """Check if the file is a raster.

    Args:
        src_filepath: input file path

    Returns:
        True if the provided file is a raster
    """
    try:
        Info(src_filepath)
        return True
    except (errors.RasterioIOError, errors.CRSError):
        pass
    return False


def convert_to_cog(
    local_filename: str,
    keep_cog_dir: str = "",
) -> str:
    """Convert raster to COG in a tmp directory.

    Args:
        local_filename: input file path
        keep_cog_dir: path to the directory to keep COG files (not used when "")

    Returns:
        file path of the converted raster
    """
    if keep_cog_dir:
        tmpcog = os.path.join(keep_cog_dir, os.path.basename(local_filename))
        if not os.path.exists(keep_cog_dir):
            os.makedirs(keep_cog_dir)
    else:
        tmpcog = os.path.join(
            os.path.dirname(local_filename), "TMPCOG", os.path.basename(local_filename)
        )
        if not os.path.exists(os.path.dirname(tmpcog)):
            os.makedirs(os.path.dirname(tmpcog))
    if os.path.exists(tmpcog) and keep_cog_dir:
        return tmpcog
    raster2cog(local_filename, tmpcog)
    return tmpcog


def apply_proj_extension(asset: Asset):
    """Apply projection extension.

    Args:
        asset: cog raster asset

    Returns:
        None
    """
    # Projection
    proj_ext = AssetProjectionExtension.ext(asset, add_if_missing=True)
    try:
        info = Info(raster_file=asset.href)
    except NoSpatialLayerException:
        logger.warning("Failed to retrieve spatial info for %s", asset.href)
    proj_ext_args = info.proj_ext_info
    logger.debug("Projection extension args: %s", proj_ext_args)
    if proj_ext_args is not None:
        logger.debug("Applying projection extension")
        proj_ext.apply(**proj_ext_args)


def get_args_for_raster_ext(asset_path: str):
    """This function returns the arguments, as a dict, for the raster extension.

    Args:
        asset_path: asset path

    Returns:
        Dict of arguments for the `RasterExtension.ext(...).apply()`
        function.
    """
    try:
        raster_info = Info(asset_path)
        bands = []
        for band in raster_info.bands:
            md, stats = raster_info.band_info(band=band)

            raster_stats = Statistics.create(**stats)
            bands.append(RasterBand.create(statistics=raster_stats, **md))
        return {"bands": bands}
    except (rasterio.errors.RasterioIOError, rasterio.errors.CRSError) as err:
        logger.warning(
            "Failed to retrieve raster information for %s (%s). "
            "Maybe it is expected because the file is not a raster.",
            err,
            asset_path,
        )
    return None


def _merge_raster_bands(b1: RasterBand, b2: RasterBand) -> RasterBand:
    """Merge two RasterBand instances.

    Args:
        b1: band 1
        b2: band 2

    Returns:
        Merged RasterBand instance
    """
    args = {}
    for band in [b1, b2]:
        for argn in band.to_dict().keys():
            args[argn] = getattr(band, argn)
    return RasterBand.create(**args)


def apply_raster_extension(asset: Asset):
    """Apply raster extension.

    Args:
        asset: cog raster asset

    Returns:
        None
    """
    # Raster
    # For the raster extension, we merge RasterBand arguments from
    # (1) what is implemented in the `task`, and (2) stacflow built-in
    # properties (stats, nodata, etc)
    raster_ext_args = get_args_for_raster_ext(asset.href)
    logger.debug("Built-in raster extension args: %s", raster_ext_args)
    if raster_ext_args is not None:
        logger.debug("Applying raster extension")
        try:
            raster_ext = RasterExtension.ext(asset)
            if raster_ext.bands:
                builtin_bands = raster_ext_args["bands"]

                # When a raster extension has already been applied to
                # the asset, we want to keep the information that is
                # already there.
                logger.debug(
                    "Found an existing raster extension for this asset. Merging %s and %s",
                    [band.to_dict() for band in builtin_bands],
                    [band.to_dict() for band in raster_ext.bands],
                )
                raster_ext_args = {
                    "bands": [
                        _merge_raster_bands(b1, b2)
                        for b1, b2 in zip(
                            builtin_bands,  # built-in
                            raster_ext.bands,  # in `task`
                        )
                    ]
                }
                logger.debug("Result: %s", raster_ext_args)
            else:
                logger.debug("No raster extension properties for asset")
        except ExtensionNotImplemented:
            logger.debug("No existing raster extension found. Creating.")
            raster_ext = RasterExtension.ext(asset, add_if_missing=True)

        raster_ext.apply(**raster_ext_args)


def apply_published_extension(item: Item):
    """Apply published date extension.

    Args:
        asset: cog raster asset

    Returns:
        None
    """
    # Timestamps
    logger.debug("Apply timestamp extension for published date")
    ts_ext = ItemTimestampsExtension.ext(item, add_if_missing=True)
    ts_ext.apply(published=datetime.now())



"""HTTP connections with various methods."""

from typing import Dict, Any
from ast import literal_eval
from pydantic import BaseModel, ConfigDict
from .logger import get_logger_for
from .utils import create_session
from .oauth2 import OAuth2Session, retrieve_token_endpoint
from .model import ApiKey
from .settings import ENV


log = get_logger_for(__name__)


class BareConnectionMethod(BaseModel):
    """Bare connection method, no extra headers."""

    model_config = ConfigDict(arbitrary_types_allowed=True)
    endpoint: str = ENV.tld_signing_endpoint

    def get_headers(self) -> Dict[str, str]:
        """Get the headers."""
        return {}


class OAuth2ConnectionMethod(BareConnectionMethod):
    """OAuth2 connection method."""

    oauth2_session: OAuth2Session = OAuth2Session()

    def get_headers(self):
        """Return the headers."""
        return {"authorization": f"bearer {self.oauth2_session.get_access_token()}"}

    def get_userinfo(self):
        """Override parent method from BareConnectionMethod."""
        openapi_url = retrieve_token_endpoint().replace("/token", "/userinfo")
        return (
            create_session()
            .get(openapi_url, timeout=10, headers=self.get_headers())
            .json()
        )


class ApiKeyConnectionMethod(BareConnectionMethod):
    """API key connection method."""

    api_key: ApiKey

    def get_headers(self):
        """Return the headers."""
        return self.api_key.to_dict()


class HTTPSession:
    """HTTP session class."""

    def __init__(self, timeout=10):
        """Initialize the HTTP session."""
        self.session = create_session()
        self.timeout = timeout
        self.headers = {
            "Content-Type": "application/json",
            "Accept": "application/json",
        }
        self._method = None

    def get_method(self):
        """Get method."""
        log.debug("Get method")
        if not self._method:
            # Lazy instantiation
            self.prepare_connection_method()
        return self._method

    def prepare_connection_method(self):
        """Set the connection method."""
        # Custom server without authentication method
        if ENV.tld_disable_auth:
            self._method = BareConnectionMethod(endpoint=ENV.tld_signing_endpoint)

        # API key method
        elif api_key := ApiKey.grab():
            self._method = ApiKeyConnectionMethod(api_key=api_key)

        # OAuth2 method
        else:
            self._method = OAuth2ConnectionMethod()

    def post(self, route: str, params: Dict):
        """Perform a POST request."""
        method = self.get_method()
        url = f"{method.endpoint}{route}"
        headers = {**self.headers, **method.get_headers()}
        log.debug("POST to %s", url)
        response = self.session.post(url, json=params, headers=headers, timeout=10)
        try:
            response.raise_for_status()
        except Exception as e:
            log.error(literal_eval(response.text))
            raise e

        return response


session = HTTPSession()


def get_headers() -> dict[str, Any]:
    """Return the headers needed to authenticate on the system."""
    return session.get_method().get_headers()


def get_userinfo() -> dict[str, str]:
    """Return userinfo."""
    return OAuth2ConnectionMethod().get_userinfo()


def get_username() -> str:
    """Return username."""
    user_info = get_userinfo()
    assert user_info, "Could not fetch user info"
    return user_info["preferred_username"]



"""NDVI loss example with pyotb."""

import pystac_client
import stackstac
from matplotlib import pyplot as plt

import teledetection

aoi_bounds = [3.944092, 43.526638, 4.014816, 43.568420]

# retrieving the relevant STAC Items
api = pystac_client.Client.open(
    "https://api.stac.teledetection.fr",
    modifier=teledetection.sign_inplace,
)

TIME_RANGE = "2016-04-10/2017-11-01"
search = api.search(
    collections=["sentinel2-l2a-theia"],
    datetime=TIME_RANGE,
    bbox=aoi_bounds,
    query={"sat:relative_orbit": {"eq": 8}, "eo:cloud_cover": {"lt": 50}},
    sortby="datetime",
    max_items=6,
)
items = search.item_collection()
print(f"{len(items)} items found")

time_steps_pc = len(items)

bands = ["B03", "B04", "B08", "CLM_R1"]
FILL_VALUE = 2**16 - 1
asset = items.items[0].assets["B03"]
proj_code = asset.ext.proj.code or ""
epsg = int(proj_code.replace("EPSG:", ""))
array = stackstac.stack(
    items,
    assets=bands,
    resolution=10,
    fill_value=1,
    bounds_latlon=aoi_bounds,
    epsg=epsg,
    chunksize=(time_steps_pc, 1, "auto", "auto"),
)
array.drop_duplicates("time")

CLM_R1 = array.sel(band="CLM_R1").squeeze().astype("uint8")
array = array.sel(band=~(array.band == "CLM_R1"))
array = array.assign_coords(CLM_R1=(("time", "y", "x"), CLM_R1.values))

source = array.sel(band=["B08", "B04", "B03"])

rgb = source.isel(time=range(6))
rgb.plot.imshow(col_wrap=3, col="time", rgb="band", vmax=2500, size=4)

nir, red = (
    source.sel(band="B08").astype("float"),
    source.sel(band="B04").astype("float"),
)
ndvi = (nir - red) / (nir + red)

ndvi.isel(time=range(6)).plot.imshow(col_wrap=3, col="time", size=4, cmap="RdYlGn")
plt.show()
